/*
 * menu.c: Web video plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <stdlib.h>
#include <time.h>
#include <vdr/skins.h>
#include <vdr/tools.h>
#include <vdr/i18n.h>
#include <vdr/osdbase.h>
#include <vdr/skins.h>
#include <vdr/font.h>
#include <vdr/osd.h>
#include "menu.h"
#include "download.h"
#include "config.h"
#include "common.h"
#include "history.h"

cCharSetConv csc = cCharSetConv("UTF-8", cCharSetConv::SystemCharacterTable());

// --- cXMLMenu --------------------------------------------------

cXMLMenu::cXMLMenu(const char *Title, int c0, int c1, int c2,
				       int c3, int c4)
: cOsdMenu(Title, c0, c1, c2, c3, c4)
{
}

bool cXMLMenu::Deserialize(const char *xml) {
  xmlDocPtr doc = xmlParseMemory(xml, strlen(xml));
  if (!doc) {
    xmlErrorPtr xmlerr =  xmlGetLastError();
    if (xmlerr) {
      error("libxml error: %s", xmlerr->message);
    }

    return false;
  }

  xmlNodePtr node = xmlDocGetRootElement(doc);
  if (node)
    node = node->xmlChildrenNode;

  while (node) {
    if (node->type == XML_ELEMENT_NODE) {
      if (!CreateItemFromTag(doc, node)) {
        warning("Failed to parse menu tag: %s", (char *)node->name);
      }
    }
    node = node->next;
  }

  xmlFreeDoc(doc);
  return true;
}

int cXMLMenu::Load(const char *xmlstr) {
  Clear();
  Deserialize(xmlstr);

  return 0;
}


// --- cNavigationMenu -----------------------------------------------------

cNavigationMenu::cNavigationMenu(cNavigationMenu **parentMenuPointer, 
				 cHistory *History, cWebviThread *dlthread)
  : cXMLMenu("", 25)
{
  history = History;
  downloadthread = dlthread;
  pluginMenuPointer = parentMenuPointer;
  statusscreen = NULL;
  UpdateHelp();
}

cNavigationMenu::~cNavigationMenu() {
  *pluginMenuPointer = NULL;
  Clear();
}

bool cNavigationMenu::CreateItemFromTag(xmlDocPtr doc, xmlNodePtr node) {
  if (!xmlStrcmp(node->name, BAD_CAST "link")) {
    NewLinkItem(doc, node);
    return true;
  } else if (!xmlStrcmp(node->name, BAD_CAST "textfield")) {
    NewTextField(doc, node);
    return true;
  } else if (!xmlStrcmp(node->name, BAD_CAST "itemlist")) {
    NewItemList(doc, node);
    return true;
  } else if (!xmlStrcmp(node->name, BAD_CAST "textarea")) {
    NewTextArea(doc, node);
    return true;
  } else if (!xmlStrcmp(node->name, BAD_CAST "button")) {
    NewButton(doc, node);
    return true;
  } else if (!xmlStrcmp(node->name, BAD_CAST "title")) {
    NewTitle(doc, node);
    return true;
  }

  return false;
}

void cNavigationMenu::AddLinkItem(cOsdItem *item,
                                  cLinkBase *ref, 
                                  cLinkBase *streamref) {
  Add(item);

  if (ref)
    links.Append(ref);
  else
    links.Append(NULL);

  if (streamref)
    streams.Append(streamref);
  else
    streams.Append(NULL);
}

void cNavigationMenu::NewLinkItem(xmlDocPtr doc, xmlNodePtr node) {
  // label, ref and object tags
  xmlChar *itemtitle = NULL, *ref = NULL, *streamref = NULL;

  node = node->xmlChildrenNode;
  while (node) {
    if (!xmlStrcmp(node->name, BAD_CAST "label")) {
      if (itemtitle)
        xmlFree(itemtitle);
      itemtitle = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    } else if (!xmlStrcmp(node->name, BAD_CAST "ref")) {
      if (ref)
        xmlFree(ref);
      ref = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    } else if (!xmlStrcmp(node->name, BAD_CAST "stream")) {
      if (streamref)
        xmlFree(streamref);
      streamref = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    }
    node = node->next;
  }
  if (!itemtitle)
    itemtitle = xmlCharStrdup("???");

  const char *titleconv = csc.Convert((char *)itemtitle);
  cOsdItem *item = new cOsdItem(titleconv);
  cSimpleLink *objlinkdata = NULL;
  cSimpleLink *linkdata = NULL;
  if (ref)
    linkdata = new cSimpleLink((char *)ref);
  if (streamref) {
    // media object
    objlinkdata = new cSimpleLink((char *)streamref);
  } else {
    // navigation link
    char *bracketed = (char *)malloc((strlen(titleconv)+3)*sizeof(char));
    if (bracketed) {
      bracketed[0] = '\0';
      strcat(bracketed, "[");
      strcat(bracketed, titleconv);
      strcat(bracketed, "]");
      item->SetText(bracketed, false);
    }
  }
  AddLinkItem(item, linkdata, objlinkdata);

  xmlFree(itemtitle);
  if (ref)
    xmlFree(ref);
  if (streamref)
    xmlFree(streamref);
}

void cNavigationMenu::NewTextField(xmlDocPtr doc, xmlNodePtr node) {
  // name attribute
  xmlChar *name = xmlGetProp(node, BAD_CAST "name");
  cHistoryObject *curhistpage = history->Current();

  // label tag
  xmlChar *text = NULL;
  node = node->xmlChildrenNode;
  while (node) {
    if (!xmlStrcmp(node->name, BAD_CAST "label")) {
      if (text)
        xmlFree(text);
      text = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    }
    node = node->next;
  }
  if (!text)
    text = xmlCharStrdup("???");

  cTextFieldData *data = curhistpage->GetTextFieldData((char *)name);
  cMenuEditStrItem *item = new cMenuEditStrItem(csc.Convert((char *)text), 
                                                data->GetValue(), 
                                                data->GetLength());
  AddLinkItem(item, NULL, NULL);

  free(text);
  if (name)
    xmlFree(name);
}

void cNavigationMenu::NewItemList(xmlDocPtr doc, xmlNodePtr node) {
  // name attribute
  xmlChar *name = xmlGetProp(node, BAD_CAST "name");
  cHistoryObject *curhistpage = history->Current();

  // label and item tags
  xmlChar *text = NULL;
  cStringList items;
  cStringList itemvalues;
  node = node->xmlChildrenNode;
  while (node) {
    if (!xmlStrcmp(node->name, BAD_CAST "label")) {
      if (text)
        xmlFree(text);
      text = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    } else if (!xmlStrcmp(node->name, BAD_CAST "item")) {
      xmlChar *str = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
      if (!str)
        str = xmlCharStrdup("???");
      xmlChar *strvalue = xmlGetProp(node, BAD_CAST "value");
      if (!strvalue)
        strvalue = xmlCharStrdup("");

      items.Append(strdup((char *)str));
      itemvalues.Append(strdup((char *)strvalue));

      xmlFree(str);
      xmlFree(strvalue);
    }
    node = node->next;
  }
  if (!text)
    text = xmlCharStrdup("???");

  cItemListData *data = curhistpage->GetItemListData((const char *)name,
                                                     items,
                                                     itemvalues);

  cMenuEditStraItem *item = new cMenuEditStraItem(csc.Convert((char *)text), 
                                                  data->GetValuePtr(), 
                                                  data->GetNumStrings(), 
                                                  data->GetStrings());
  AddLinkItem(item, NULL, NULL);

  xmlFree(text);
  if (name)
    xmlFree(name);
}

void cNavigationMenu::NewTextArea(xmlDocPtr doc, xmlNodePtr node) {
  // label tag
  xmlChar *itemtitle = NULL;
  node = node->xmlChildrenNode;
  while (node) {
    if (!xmlStrcmp(node->name, BAD_CAST "label")) {
      if (itemtitle)
        xmlFree(itemtitle);
      itemtitle = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    }
    node = node->next;
  }
  if (!itemtitle)
    return;

  const cFont *font = cFont::GetFont(fontOsd);
  cTextWrapper tw(csc.Convert((char *)itemtitle), font, cOsd::OsdWidth());
  for (int i=0; i < tw.Lines(); i++) {
    AddLinkItem(new cOsdItem(tw.GetLine(i), osUnknown, false), NULL, NULL);
  }

  xmlFree(itemtitle);
}

void cNavigationMenu::NewButton(xmlDocPtr doc, xmlNodePtr node) {
  // label and submission tags
  xmlChar *itemtitle = NULL, *submission = NULL;
  cHistoryObject *curhistpage = history->Current();

  node = node->xmlChildrenNode;
  while (node) {
    if (!xmlStrcmp(node->name, BAD_CAST "label")) {
      if (itemtitle)
        xmlFree(itemtitle);
      itemtitle = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    } else if (!xmlStrcmp(node->name, BAD_CAST "submission")) {
      if (submission)
        xmlFree(submission);
      submission = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    }
    node = node->next;
  }
  if (!itemtitle)
    itemtitle = xmlCharStrdup("???");

  cSubmissionButtonData *data = \
          new cSubmissionButtonData((char *)submission, curhistpage);
  const char *titleconv = csc.Convert((char *)itemtitle); // do not free
  char *newtitle = (char *)malloc((strlen(titleconv)+3)*sizeof(char));
  if (newtitle) {
    newtitle[0] = '\0';
    strcat(newtitle, "[");
    strcat(newtitle, titleconv);
    strcat(newtitle, "]");

    cOsdItem *item = new cOsdItem(newtitle);
    AddLinkItem(item, data, NULL);
    free(newtitle);
  }

  xmlFree(itemtitle);
  if (submission)
    xmlFree(submission);
}

void cNavigationMenu::NewTitle(xmlDocPtr doc, xmlNodePtr node) {
  xmlChar *newtitle = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
  if (newtitle) {
    SetTitle(csc.Convert((char *)newtitle));
    xmlFree(newtitle);
  }
}

eOSState cNavigationMenu::ProcessKey(eKeys Key)
{
  int old = Current();
  eOSState state = cXMLMenu::ProcessKey(Key);

  //debug("cNavigationMenu::ProcessKey: current = %d, size = %d", Current(), links.Size());
  if ((Current() < 0) || (Current() >= links.Size()))
    return state;

  if (state == osUnknown) {
    switch (Key) {
    case kInfo:
      // The alternative link is active only when object links are
      // present.
      if (streams.At(Current()))
	state = Select(links.At(Current()), LT_REGULAR);
      break;
    case kOk:
      // Primary action: download media object or, if not a media
      // link, follow the navigation link.
      if (streams.At(Current()))
	state = Select(streams.At(Current()), LT_MEDIA);
      else
	state = Select(links.At(Current()), LT_REGULAR);
      break;
    case kBlue:
      // Secondary action: start streaming if a media object
      if (streams.At(Current())) {
	state = Select(streams.At(Current()), LT_STREAMINGMEDIA);
      }
      break;
    case kRed:
      state = HistoryBack();
      break;
    case kGreen:
      state = HistoryForward();
      break;
    case kYellow:
      SetStatusScreen(new cStatusScreen(this, downloadthread));
      state = AddSubMenu(statusscreen);
      break;
    default:
      break;
    }
  } else {
    // If the key press caused the selected item to change, we need to
    // update the help texts.
    //
    // In cMenuEditStrItem key == kOk with state == osContinue
    // indicates leaving the edit mode. We want to update the help
    // texts in this case also.
    if ((old != Current()) || 
	((Key == kOk) && (state == osContinue))) {
      UpdateHelp();
    }
  }

  //debug("cNavigationMenu::ProcessKey, state = %d", state);
  return state;
}

eOSState cNavigationMenu::Select(cLinkBase *link, eLinkType type)
{
  if (!link) {
    return osContinue;
  }
  char *ref = link->GetURL();
  if (!ref) {
    error("link->GetURL() == NULL in cNavigationMenu::Select");
    return osContinue;
  }

  if (type == LT_MEDIA) {
    downloadthread->AddRequest(new cFileDownloadRequest(history->Current()->GetID(),
                                                        ref, 
                                                        webvideoConfig->GetDownloadPath()));
    Skins.Message(mtInfo, tr("Downloading in the background"));
  } else if (type == LT_STREAMINGMEDIA) {
    downloadthread->AddRequest(new cStreamUrlRequest(history->Current()->GetID(),
						     ref));
    Skins.Message(mtInfo, tr("Starting player..."));
    return osEnd;
  } else {
    downloadthread->AddRequest(new cMenuRequest(history->Current()->GetID(),
                                                ref));
    Skins.Message(mtStatus, tr("Retrieving..."));
  }

  return osContinue;
}

void cNavigationMenu::Clear(void) {
  cXMLMenu::Clear();
  SetTitle("");
  for (int i=0; i < links.Size(); i++) {
    if (links[i])
      delete links[i];
    if (streams[i])
      delete streams[i];
  }
  links.Clear();
  streams.Clear();
}

void cNavigationMenu::Populate(cHistoryObject *page, const char *statusmsg) {
  Load(page->GetOSD());

  // Make sure that an item is selected (if there is at least
  // one). The help texts are not updated correctly if no item is
  // selected.

  SetCurrent(Get(page->GetSelected()));
  UpdateHelp();
  SetStatus(statusmsg);
}

eOSState cNavigationMenu::HistoryBack() {
  cHistoryObject *cur = history->Current();

  if (cur)
    cur->RememberSelected(Current());

  cHistoryObject *page = history->Back();
  if (page) {
    Populate(page);
    Display();
  }
  return osContinue;
}

eOSState cNavigationMenu::HistoryForward() {
  cHistoryObject *before = history->Current();
  cHistoryObject *after = history->Forward();

  if (before)
    before->RememberSelected(Current());

  // Update only if the menu really changed
  if (before != after) {
    Populate(after);
    Display();
  }
  return osContinue;
}

void cNavigationMenu::UpdateHelp() {
  const char *red = (history->Current() != history->First()) ? tr("Back") : NULL;
  const char *green = (history->Current() != history->Last()) ? tr("Forward") : NULL;
  const char *yellow = tr("Status");
  const char *blue = ((Current() >= 0) && (streams.At(Current()))) ? tr("Play") : NULL;

  SetHelp(red, green, yellow, blue);
}

cStatusScreen *cNavigationMenu::GetStatusScreen() {
  if (HasSubMenu()) {
    return statusscreen;
  } else {
    return NULL;
  }
}

// --- cStatusScreen -------------------------------------------------------

cStatusScreen::cStatusScreen(cNavigationMenu* parent, 
                             cWebviThread *thread)
  : cOsdMenu(tr("Unfinished downloads"), 40)
{
  parentmenu = NULL;
  downloadthread = thread;
  SetHelp(NULL, NULL, tr("Abort"), NULL);

  int charsperline = cOsd::OsdWidth() / cFont::GetFont(fontOsd)->Width('M');
  SetCols(charsperline-5);

  Update();
}

cStatusScreen::~cStatusScreen() {
  for (int i=0; i<summaries.Size(); i++) {
    delete summaries[i];
  }
  summaries.Clear();
}

void cStatusScreen::Update() {
  int c = Current();

  Clear();
  for (int i=0; i<summaries.Size(); i++) {
    delete summaries[i];
  }
  summaries.Clear();

  downloadthread->GetUnfinishedFilesSummary(&summaries);
  if (summaries.Size() == 0) {
    SetTitle(tr("No active downloads"));
  } else {
    for (int i=0; i<summaries.Size(); i++) {
      cString dltitle;
      cDownloadProgress *s = summaries[i];

      if (s->GetPercentage() < 0.0)
        dltitle = cString::sprintf("%s\t???%%", s->GetTitle());
      else
        dltitle = cString::sprintf("%s\t%3d%%", s->GetTitle(), 
                                   (int)(100*s->GetPercentage()+0.5));
      Add(new cOsdItem(dltitle));
    }

    if (c >= 0)
      SetCurrent(Get(c));
  }

  lastupdate = time(NULL);

  Display();
}

bool cStatusScreen::NeedsUpdate() {
  return (Count() > 0) && (time(NULL) - lastupdate >= updateInterval);
}

eOSState cStatusScreen::ProcessKey(eKeys Key) {
  eOSState state = cOsdMenu::ProcessKey(Key);

  if (state == osUnknown) {
    switch (Key) {
    case kYellow:
      if ((Current() >= 0) && (Current() < summaries.Size())) {
        downloadthread->Abort(summaries[Current()]->GetDownloadID());
	delete summaries[Current()];
        summaries.Remove(Current());
        Del(Current());
        Display();
      }
      return osContinue;
    default:
      break;
    }
  }

  return state;
}
